import logging
import h5py
import numpy as np
from pathlib import Path
from tqdm import tqdm
from tensornetwork import Node, conj
from tensornetwork.matrixproductstates.finite_mps import FiniteMPS
# from tnpy.linalg import svd
from tnpy.operators import MPO
from typing import List, Union


class FiniteAlgorithmBase:

    mps_cls = FiniteMPS

    def __init__(self, mpo: MPO, chi: Union[int, None] = None, init_method: str = 'random'):
        """

        Args:
            mpo:
            chi:
            init_method:
        """
        if init_method == 'random':
            assert isinstance(chi, int), "chi has to be given for random init"
        self.mpo = mpo
        self.d = mpo.physical_dimensions
        self.chi = chi
        self.left_envs = {}
        self.right_envs = {}
        self.left_norms = {}
        self.right_norms = {}
        self._mps = None
        self.mps = init_method

    def close(self):
        logging.info("Deleting left-/right- environments and norms")
        del self.left_envs
        del self.right_envs
        del self.left_norms
        del self.right_norms

    @property
    def N(self) -> int:
        return len(self.mpo.nodes)

    @property
    def dtype(self):
        return self.mpo[0].tensor.dtype

    @property
    def bond_dimensions(self):
        return self._mps.bond_dimensions

    @staticmethod
    def generate_bond_dimensions(N: int, d: int, chi: int) -> List[int]:
        if N % 2 == 0:
            D = [min(d ** i, chi) for i in range(1, N // 2)]
            D += [int(min(d ** (N / 2), chi))] + D[::-1]
        else:
            D = [min(d ** i, chi) for i in range(1, (N + 1) // 2)]
            D += D[::-1]
        return D

    @property
    def mps(self) -> FiniteMPS:
        return self._mps

    @mps.setter
    def mps(self, init_method: str = 'random') -> None:
        if init_method == "random":
            logging.info("Initializing finite MPS randomly")
            D = FiniteAlgorithmBase.generate_bond_dimensions(self.N, self.d, self.chi)
            self._mps = self.mps_cls.random([self.d]*self.N, D, self.dtype)
            # self.normalize_mps(direction=1, normalize_sv=True)
        elif Path(init_method).suffix:
            self.load_mps(init_method)
        else:
            raise KeyError("Invalid init method or file does not exist")
        self._init_envs()

    def load_mps(self, filename: str) -> None:
        logging.info(f"Initializing finite MPS from file: {filename}")
        path = Path(filename)
        if path.suffix == '.hdf5':
            # TODO: Not Implemented
            NotImplemented
        elif path.suffix == '.npz':
            npzfile = np.load(path)
            # @TODO: this has to be converted into object FiniteMPS
            self._mps = [npzfile['arr_%i' % i] for i in range(len(npzfile.files))]
            npzfile.close()
        else:
            raise FileNotFoundError(f"File {path} not exists or not supported.")

    def save_mps(self, filename: str) -> None:
        path = Path(filename)
        tensor_dataset = map(lambda node: node.tensor, self._mps.nodes)
        if path.suffix == '.hdf5':
            # @TODO: this requires more cares on axes
            with h5py.File(path, 'w') as hf:
                hf.create_dataset(str(path), tensor_dataset)
                hf.close()
        elif path.suffix == '.npz':
            np.savez(path, *tensor_dataset)
        else:
            raise NameError(f"File extension {path.suffix} is not supported.")

    def mps_shape(self, site: int):
        return self._mps.get_tensor(site).shape

    # def normalize_mps(self, direction, normalize_sv=False):
    #     if direction == 1:
    #         iterator = range(self.N-1)
    #     elif direction == -1:
    #         iterator = range(self.N-1, 0, -1)
    #
    #     for site in iterator:
    #         theta = self._mps.nodes[site].tensor
    #         if direction == 1:
    #             theta = theta.reshape(self.d * self.mps_shape(site)[0], -1)
    #         elif direction == -1:
    #             theta = theta.reshape(-1, self.d * self.mps_shape(site)[2])
    #         u, s, vt = np.linalg.svd(theta, full_matrices=False)
    #         if normalize_sv:
    #             s /= np.linalg.norm(s)
    #         if direction == 1:
    #             self._mps.nodes[site] = Node(u.reshape(self.mps_shape(site)))
    #             residual = Node(np.dot(np.diagflat(s), vt))
    #             G = self._mps.nodes[site+1]
    #             residual[1] ^ G[0]
    #             self._mps.nodes[site+1] = residual @ G
    #             self._update_left_env(site+1)
    #         elif direction == -1:
    #             self._mps.nodes[site] = Node(vt.reshape(self.mps_shape(site)))
    #             residual = Node(np.dot(u, np.diagflat(s)))
    #             G = self._mps.nodes[site-1]
    #             G[2] ^ residual[0]
    #             self._mps.nodes[site-1] = G @ residual
    #             self._update_right_env(site-1)

    def _init_envs(self):
        # @TODO: only need to do for one direction
        logging.info("Initializing left environments")
        for site in tqdm(range(1, self.N)):
            self._update_left_env(site)
        logging.info("Initializing right environments")
        for site in tqdm(range(self.N-2, -1, -1)):
            self._update_right_env(site)

    def _init_norms(self):
        # @TODO: only need to do for one direction
        logging.info("Initializing left norms")
        for site in tqdm(range(1, self.N)):
            self._update_left_norm(site)
        logging.info("Initializing right norms")
        for site in tqdm(range(self.N-2, -1, -1)):
            self._update_right_norm(site)

    def _update_left_env(self, site):
        W = self.mpo[site-1]
        M = Node(self._mps.get_tensor(site-1))
        M_conj = Node(conj(self._mps.get_tensor(site-1)))
        if site == 1:
            M[0] ^ M_conj[0]
            M[1] ^ W[1]
            W[2] ^ M_conj[1]
            self.left_envs[site] = M @ W @ M_conj
        else:
            L = self.left_envs[site-1]
            L[0] ^ M[0]
            L[1] ^ W[0]
            L[2] ^ M_conj[0]
            M[1] ^ W[2]
            M_conj[1] ^ W[3]
            self.left_envs[site] = L @ M @ W @ M_conj

    def _update_right_env(self, site):
        W = self.mpo[site+1]
        M = Node(self._mps.get_tensor(site+1))
        M_conj = Node(conj(self._mps.get_tensor(site+1)))
        if site == self.N-2:
            M[2] ^ M_conj[2]
            M[1] ^ W[1]
            W[2] ^ M_conj[1]
            self.right_envs[site] = M @ W @ M_conj
        else:
            R = self.right_envs[site+1]
            R[0] ^ M[2]
            R[1] ^ W[1]
            R[2] ^ M_conj[2]
            M[1] ^ W[2]
            M_conj[1] ^ W[3]
            self.right_envs[site] = R @ M @ W @ M_conj

    def _update_left_norm(self, site):
        M = self._mps.nodes[site-1]
        M_conj = conj(self._mps.nodes[site-1])
        if site == 1:
            M[0] ^ M_conj[0]
            M[1] ^ M_conj[1]
            self.left_norms[site] = M @ M_conj
        else:
            L = self.left_norms[site-1]
            L[0] ^ M[0]
            L[1] ^ M_conj[0]
            M[1] ^ M_conj[1]
            self.left_norms[site] = L @ M @ M_conj

    def _update_right_norm(self, site):
        M = self._mps.nodes[site+1]
        M_conj = conj(self._mps.nodes[site+1])
        if site == self.N-2:
            M[2] ^ M_conj[2]
            M[1] ^ M_conj[1]
            self.right_norms[site] = M @ M_conj
        else:
            R = self.right_norms[site+1]
            R[0] ^ M[2]
            R[1] ^ M_conj[2]
            M[1] ^ M_conj[1]
            self.right_norms[site] = R @ M @ M_conj



import time
import logging
import numpy as np
from tensornetwork import Node
from itertools import count
from tnpy.finite_algorithm_base import FiniteAlgorithmBase
from tnpy.linalg import svd, eigshmv
from tnpy.operators import MPO
from typing import Iterable, Union, Tuple


class FiniteDMRG(FiniteAlgorithmBase):

    def __init__(self, mpo: MPO, chi: Union[int, None], init_method='random'):
        """

        Args:
            mpo:
            chi: Maximum bond dimension of MPS
            init_method: 'random' or a filepath
        """
        super(FiniteDMRG, self).__init__(mpo, chi, init_method)

    def _unit_solver(self, site, tol=1e-7) -> Tuple[float, np.ndarray]:
        W = self.mpo[site]

        def matvec(x):
            M = Node(x.reshape(self.mps_shape(site)))
            if site == 0:
                R = self.right_envs[site]
                R[0] ^ M[2]
                R[1] ^ W[0]
                M[1] ^ W[1]
                result = M @ W @ R
            elif site == self.N-1:
                L = self.left_envs[site]
                L[0] ^ M[0]
                L[1] ^ W[0]
                M[1] ^ W[1]
                result = L @ M @ W
            else:
                L = self.left_envs[site]
                R = self.right_envs[site]
                L[0] ^ M[0]
                L[1] ^ W[0]
                R[0] ^ M[2]
                R[1] ^ W[1]
                M[1] ^ W[2]
                result = L @ M @ W @ R
            return result.tensor.reshape(x.shape)
        v0 = self._mps.get_tensor(site).reshape(-1, 1)
        return eigshmv(matvec, v0, tol=0.1*tol)

    def _modified_density_matrix(self, site, alpha=0):
        # TODO: return dm
        NotImplemented

    def sweep(self, iterator: Iterable, tol: float = 1e-7) -> float:
        """

        Args:
            iterator:
            tol:

        Returns:

        """
        direction = 1 if iterator[0] < iterator[-1] else -1
        for site in iterator:
            E, theta = self._unit_solver(site, tol)
            logging.info(f"Sweeping to site [{site+1}/{self.N}], E/N = {E/self.N}")
            if direction == 1:
                theta = theta.reshape(self.d * self.mps_shape(site)[0], -1)
            elif direction == -1:
                theta = theta.reshape(-1, self.d * self.mps_shape(site)[2])
            u, s, vt = svd(theta, chi=self.mps_shape(site)[1+direction])
            if direction == 1:
                self._mps.tensors[site] = u.reshape(self.mps_shape(site))
                residual = Node(np.dot(np.diagflat(s), vt))
                M = Node(self._mps.get_tensor(site+1))
                residual[1] ^ M[0]
                self._mps.tensors[site+1] = (residual @ M).tensor
                self._update_left_env(site+1)
            elif direction == -1:
                self._mps.tensors[site] = vt.reshape(self.mps_shape(site))
                residual = Node(np.dot(u, np.diagflat(s)))
                M = Node(self._mps.get_tensor(site-1))
                M[2] ^ residual[0]
                self._mps.tensors[site-1] = (M @ residual).tensor
                self._update_right_env(site-1)
        return E

    def update(self, tol: float = 1e-7, max_sweep: int = 100):
        """

        Args:
            tol:
            max_sweep:

        Returns:

        """
        logging.info(f"Set up tol = {tol}, up to maximally {max_sweep} sweeps")
        clock = [time.process_time()]
        for n_sweep in count(start=1):
            logging.info(f"In sweep epoch [{n_sweep}/{max_sweep}]")
            El = self.sweep(range(self.N-1))
            Er = self.sweep(range(self.N-1, 0, -1))
            clock.append(time.process_time()-clock[-1])
            dE = (El - Er)/self.N
            if abs(dE) < tol:
                break
            elif n_sweep == max_sweep:
                # @TODO: dump mps to file and raise error
                logging.warning(f"Maximum number of sweeps {max_sweep} reached, "
                                f"yet dE/N = {dE} > tol = {tol}")
                break
            elif abs(dE) > tol and dE < 0:
                raise ValueError(f"Fail on lowering energy, got dE/N = {dE}")
        logging.info(f"{n_sweep} loops, best of 3: {np.mean(np.sort(clock)[:3])} sec per loop")


class Projector:

    def __init__(self):
        NotImplemented



import numpy as np
from tnpy.operators import MPO, FullHamiltonian
from typing import Tuple


class ExactDiagonalization(FullHamiltonian):

    def __init__(self, mpo: MPO):
        super(ExactDiagonalization, self).__init__(mpo)
        self._evals, self._evecs = self._eigen_solver()

    @property
    def evals(self) -> np.ndarray:
        return self._evals

    @property
    def evecs(self) -> np.ndarray:
        return self._evecs

    def _eigen_solver(self) -> Tuple[np.ndarray, np.ndarray]:
        return np.linalg.eigh(self.matrix)



import numpy as np
from tnpy.model import ModelBase
from tnpy.operators import SpinOperators


class XXZ(ModelBase):

    def __init__(self, N: int, delta: float):
        """

        Args:
            N: System size.
            delta: Coupling strength on z direction.
        """
        super(XXZ, self).__init__(N)
        self.delta = delta

    def _elem(self, site: int) -> np.ndarray:
        Sp, Sm, Sz, I2, O2 = SpinOperators()
        return np.array(
            [[I2, -0.5 * Sp, -0.5 * Sm, -self.delta * Sz, O2],
             [O2, O2, O2, O2, Sm],
             [O2, O2, O2, O2, Sp],
             [O2, O2, O2, O2, Sz],
             [O2, O2, O2, O2, I2]]
        )



import numpy as np
import primme
from scipy.sparse.linalg import LinearOperator
from typing import Callable


def svd(A: np.ndarray, chi: int):
    """

    Args:
        A: Input Matrix
        chi: Truncated dimension

    Returns:

    """
    u, s, vt = np.linalg.svd(A, full_matrices=False)
    u = u[:, 0:chi]
    s = s[0:chi]
    vt = vt[0:chi, :]
    return u, s, vt


def qr(A: np.ndarray, chi: int):
    q, r = np.linalg.qr(A, mode='reduced')
    q = q[:, 0:chi]
    r = r[0:chi, :]
    return q, r


def eigshmv(Afunc: Callable, v0: np.ndarray, tol=0):
    """

    Args:
        Afunc:
        v0:
        tol:

    Returns:

    """
    A = LinearOperator((v0.shape[0], v0.shape[0]), matvec=Afunc)
    evals, evecs = primme.eigsh(A, k=1, which='SA', v0=v0, tol=tol)
    return evals[0], evecs


class KrylovExpm:

    def __init__(self, delta, mat, v0, n=20):
        self.delta = delta
        self.mat = mat
        self.v0 = v0 / np.linalg.norm(v0)
        self.n = n
        assert(n <= mat.shape[0])

    def orthonormalize(self, vecs):
        pass

    def construct_krylov_space(self):
        vecs = [self.v0]
        for i in range(self.n-1):
            vecs.append(np.dot(self.mat, vecs[-1]))
        # vecs = self.orthonormalize(vecs)
        print(len(vecs))



import numpy as np
from tnpy.model import ModelBase
from tnpy.operators import SpinOperators


class Thirring(ModelBase):

    def __init__(self, N: int, g: float, ma: float, lamda: float, s_target: int) -> None:
        """

        Args:
            N: System size.
            g: Bare coupling g.
            ma: Bare mass.
            lamda: Penalty strength (of Lagrangian multiplier).
            s_target: The targeting total Sz charge sector.
        """
        super(Thirring, self).__init__(N)
        self.g = g
        self.ma = ma
        self.lamda = lamda
        self.s_target = s_target

    def _elem(self, site: int) -> np.ndarray:
        Sp, Sm, Sz, I2, O2 = SpinOperators()

        beta = self.g + ((-1.0) ** site * self.ma) - 2.0 * self.lamda * self.s_target
        gamma = self.lamda * (0.25 + self.s_target ** 2 / self.N) + 0.25 * self.g

        return np.array(
            [[I2, -0.5 * Sp, -0.5 * Sm, 2.0 * np.sqrt(self.lamda) * Sz, self.g * Sz, gamma * I2 + beta * Sz],
             [O2, O2, O2, O2, O2, Sm],
             [O2, O2, O2, O2, O2, Sp],
             [O2, O2, O2, I2, O2, np.sqrt(self.lamda) * Sz],
             [O2, O2, O2, O2, O2, Sz],
             [O2, O2, O2, O2, O2, I2]],
            dtype=float
        )



import logging
import numpy as np
from functools import wraps
from itertools import count
from collections import deque
from copy import copy
from tensornetwork import ncon, Node, Tensor
from dataclasses import dataclass, field
from tnpy.operators import MPO
from typing import Tuple, List, Union, Callable, Iterator, Dict


@dataclass
class Projection:
    HarmonicRitz: str = 'primme_proj_harmonic'
    RefinedRitz: str = 'primme_proj_refined'
    RayleighRitz: str = 'primme_proj_RR'


@dataclass
class TreeNode:
    """
    The node object in binary tree.

    Attributes:
        id: ID of this TreeNode.
        node: Container of :class:`~tensornetwork.Node`,
            which is supposed to take the ordering of axes `(1st physical bond, 2nd physical bond, virtual bond)`.
        left:
        right:

    Notes:
        This class differs from :class:`~tensornetwork.Node`,
        which refers to the node in a tensor network.
        Please be aware of this bad naming.
    """
    id: int
    node: Node
    left: 'TreeNode' = None
    right: 'TreeNode' = None

    def __eq__(self, other: 'TreeNode') -> bool:
        """
        Check 2 TreeNodes are identical.

        Args:
            other:

        Returns:

        Warnings:
            This is a weak condition, only the id of these 2 TreeNodes
            and the id of their 1st generation children (if not a leaf) are compared.
            It's not guaranteed that the value of :attr:`~TreeNode.node` will be the same,
            as well as their posterity.
        """
        if self.is_leaf and other.is_leaf:
            return self.id == other.id
        elif self.is_leaf or other.is_leaf:
            return False
        return self.id == other.id and self.left.id == other.left.id and self.right.id == other.right.id

    @property
    def is_leaf(self) -> bool:
        return self.left is None and self.right is None

    @property
    def shape(self) -> Tuple[int]:
        return self.node.tensor.shape


def check_root(func: Callable) -> Callable:
    """
    Helper function for checking the root in :class:`~TensorTree`.
    This is supposed to be applied on the member function of :class:`~TensorTree` as a decorator.

    Args:
        func:

    Returns:

    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        assert isinstance(args[0], TensorTree), "This is supposed to be used on class member of TensorTree."
        if not args[0].has_root:
            raise RuntimeError("Cannot find root in tree.")
        return func(*args, **kwargs)
    return wrapper


class TensorTree:

    def __init__(self, leaves: List[TreeNode]):
        """
        The bottom-up binary tree where each node contains a tensor.

        Args:
            leaves:
        """
        self._leaves = {leaf.id: leaf for leaf in leaves}
        self._tree = copy(self._leaves)
        self._horizon = list(self._tree.keys())
        self._root = None
        self._top_tensor = None

    def __iter__(self) -> Iterator:
        return iter(self._tree.values())

    @property
    def n_nodes(self) -> int:
        """
        Total number of nodes in tree.

        Returns:
            n_nodes:
        """
        return len(self._tree)

    @property
    def n_layers(self) -> int:
        def max_depth(current_node: TreeNode) -> int:
            if current_node is not None:
                left_depth = max_depth(current_node.left)
                right_depth = max_depth(current_node.right)
                return max(left_depth, right_depth) + 1
            return 0
        return max_depth(self.root)

    @property
    def n_leaves(self) -> int:
        return len(self._leaves)

    @property
    def root(self) -> Union[None, TreeNode]:
        return self._root

    @property
    def has_root(self) -> bool:
        return self.root is not None

    @property
    def horizon(self) -> List[int]:
        return self._horizon

    @property
    def top_tensor(self):
        return self._top_tensor

    @top_tensor.setter
    def top_tensor(self, top_tensor):
        self._top_tensor = top_tensor

    def append(self, node: TreeNode) -> None:
        self._tree[node.id] = node
        self._horizon[self._horizon.index(node.left.id)] = node.id
        self._horizon.remove(node.right.id)
        if self.n_nodes == 2 * self.n_leaves - 1:
            self._root = node

    def node(self, node_id: int) -> TreeNode:
        return self._tree[node_id]

    def leaf(self, node_id: int) -> TreeNode:
        return self._leaves[node_id]

    @check_root
    def find_path(self, node_id: int) -> List[int]:
        """
        Find the path to targeted node starting from the root.

        Args:
            node_id:

        Returns:
            path: ID of nodes to targeted node, including targeted node itself.

        Raises:
            If path is empty, raises KeyError for input ``node_id``.

        See Also:
            The function :py:func:`TensorTree.ancestor` does the same job as this function,
            while the returning type is List[:class:`~TreeNode`] there.
        """
        def find_node(current_node: TreeNode, trial_path: List[int]) -> bool:
            """
            Find the node with depth-first search.

            Args:
                current_node: Current node in iterative depth-first searching
                trial_path: List for recording the trial path to every leaf.

            Returns:
                found: Return True if the given ``node_id`` is found in tree, else False.
            """
            if current_node is not None:
                trial_path.append(current_node.id)
                if current_node.id == node_id:
                    return True
                elif find_node(current_node.left, trial_path) or\
                        find_node(current_node.right, trial_path):
                    return True
                trial_path.pop()
            return False

        path = []
        if not find_node(self.root, path):
            raise KeyError("The given node_id is not found in tree.")
        return path

    def ancestor(self, node_id: int) -> List[TreeNode]:
        """
        Find all the ancestors of targeted node in the order starting from the root.

        Args:
            node_id:

        Returns:
            ancestor:

        See Also:
            The function :func:`~TensorTree.find_path` does the same job as this function,
            while the returning type is List[int] there.
        """
        path = self.find_path(node_id)
        return [self.node(idx) for idx in path][:-1]

    @check_root
    def common_ancestor(self, node_id1: int, node_id2: int) -> List[int]:
        """
        Find all common ancestor of two given nodes in the order starting from the root.

        Args:
            node_id1:
            node_id2:

        Returns:

        """
        path1 = self.find_path(node_id1)
        path2 = self.find_path(node_id2)
        first_unmatched = next(idx for idx, (x, y) in enumerate(zip(path1, path2)) if x != y)
        return path1[:first_unmatched]

    def filter_out_leaf(self, node_ids: List[int]) -> List[int]:
        return [node_id for node_id in node_ids if not self.node(node_id).is_leaf]

    class Contractible:
        """
        Accessory helper class for contracting TreeNodes.
        """

        def __init__(self, parent_cls: 'TensorTree', node_ids: List[int], open_bonds: List[Tuple[int, str]]):
            self.parent_cls = parent_cls
            self.open_bonds = open_bonds
            self.input_checker(node_ids, open_bonds)
            self._contractible = {
                k: v for node_id in node_ids for k, v in zip(
                    [f"{node_id}", f"conj{node_id}"],
                    [self.cell(node_id), self.cell(node_id, conj=True)]
                )
            }

        def __getitem__(self, node_id: str):
            return self._contractible[node_id]['bonds']

        @property
        def tensors(self) -> List[Tensor]:
            return [node['tensor'] for node in self._contractible.values()]

        @property
        def network_structure(self) -> List[Tuple[str, str, str]]:
            return [tuple(node['bonds']) for node in self._contractible.values()]

        @property
        def out_order(self) -> List[str]:
            edge_bonds = [self.syntax(*open_bond) for open_bond in self.open_bonds]
            conj_edge_bonds = [self.syntax(*open_bond, conj=True) for open_bond in self.open_bonds]
            return [self.syntax(self.parent_cls.root.id, 2)] + edge_bonds +\
                   [self.syntax(self.parent_cls.root.id, 2, conj=True)] + conj_edge_bonds

        @staticmethod
        def syntax(node_id: int, idx: Union[int, str], conj: bool = False) -> str:
            if type(idx) is str:
                idx = 0 if idx == 'left' else 1
            prefix = "Conj" if conj else ""
            return f"-{prefix}Node{node_id}Sub{idx}"

        def cell(self, node_id: int, conj: bool = False) -> Dict:
            def _tensor_setter() -> Tensor:
                node = self.parent_cls.node(node_id).node
                return node.copy().tensor if conj else node.tensor

            def _bonds_setter() -> List[str]:
                return [self.syntax(node_id, idx, conj) for idx in range(3)]
            return {'tensor': _tensor_setter(), 'bonds': _bonds_setter()}

        def input_checker(self, node_ids: List[int], open_bonds: List[Tuple[int, str]]):
            if len(node_ids) > len(set(node_ids)):
                raise KeyError("Every id in node_ids must be given uniquely.")
            for elem in open_bonds:
                if not isinstance(elem, tuple) and list(map(type, elem)) == [int, str]:
                    raise TypeError("Argument `open_bonds` takes the type List[Tuple[int, str]].")
                if elem[1] not in ['left', 'right']:
                    raise KeyError("Second argument of `open_bonds` must be either `left` or `right`.")
                if getattr(self.parent_cls.node(elem[0]), elem[1]).id in node_ids:
                    raise KeyError("Child node of `open_bonds` cannot present in `node_ids`.")

    @check_root
    def contract_nodes(self, node_ids: List[int], open_bonds: List[Tuple[int, str]] = None,
                       return_out_order: bool = False) -> Union[Tensor, Tuple[Tensor, List[str]]]:
        """

        Args:
            node_ids:
            open_bonds:
            return_out_order:

        Returns:
            contracted_network:
            out_order:
        """
        if open_bonds is None:
            open_bonds = []
        node_queue = deque([self.root])
        contractible = self.Contractible(self, node_ids, open_bonds)

        def update_contractible(parent: TreeNode, on: str):
            child = getattr(parent, on)
            on_bond_idx = 0 if on == 'left' else 1
            if child.id in node_ids:
                assert not child.is_leaf, "node_ids contains leaves"
                contractible[f"{parent.id}"][on_bond_idx] = f"Node{parent.id}ToNode{child.id}"
                contractible[f"{child.id}"][2] = f"Node{parent.id}ToNode{child.id}"
                contractible[f"conj{parent.id}"][on_bond_idx] = f"ConjNode{parent.id}ToConjNode{child.id}"
                contractible[f"conj{child.id}"][2] = f"ConjNode{parent.id}ToConjNode{child.id}"
                if (child.id, 'left') or (child.id, 'right') not in open_bonds:
                    node_queue.append(child)
            elif (parent.id, on) not in open_bonds:
                contractible[f"conj{parent.id}"][on_bond_idx] = f"Node{parent.id}Sub{on_bond_idx}" \
                                                                f"ToConjNode{parent.id}Sub{on_bond_idx}"
                contractible[f"{parent.id}"][on_bond_idx] = f"Node{parent.id}Sub{on_bond_idx}" \
                                                            f"ToConjNode{parent.id}Sub{on_bond_idx}"

        while len(node_queue) > 0:
            current_node = node_queue.popleft()
            update_contractible(current_node, on='left')
            update_contractible(current_node, on='right')

        out_tensor = ncon(
            contractible.tensors,
            contractible.network_structure,
            out_order=contractible.out_order
        )
        if return_out_order:
            return out_tensor, contractible.out_order
        return out_tensor

    @check_root
    def plot(self):
        NotImplemented


class TSDRG:

    @dataclass
    class GapCache:
        """
        Helper class for storing the cache of energy gap in tSDRG algorithm.
        """
        parent_cls: 'TSDRG'
        gap: List[float] = field(default_factory=list)
        evecs: List[np.ndarray] = field(default_factory=list)
        evals: np.ndarray = None

        def __post_init__(self):
            for site in range(self.parent_cls.n_nodes - 1):
                evals, evecs = self.parent_cls.eigen_solver(self.parent_cls.block_hamiltonian(site))
                self.gap.append(self.parent_cls.truncation_gap(evals))
                self.evecs.append(evecs)
                if self.parent_cls.n_nodes == 2:
                    self.evals = evals

    def __init__(self, mpo: MPO, chi: int):
        self.mpo = mpo
        self._chi = chi
        self._N = len(mpo.nodes)
        self._tree = TensorTree(
            [
                TreeNode(id=site, node=node)
                for site, node in enumerate(self.mpo)
            ]
        )
        self.gap_cache = self.GapCache(self)
        self._top_tensor = None

    @property
    def N(self) -> int:
        return self._N

    @property
    def n_nodes(self) -> int:
        return len(self.mpo.nodes)

    @property
    def chi(self) -> int:
        return self._chi

    @property
    def v_left(self) -> Node:
        return self.mpo.v_left

    @property
    def v_right(self) -> Node:
        return self.mpo.v_right

    @property
    def tree(self) -> TensorTree:
        return self._tree

    @property
    def top_tensor(self) -> Tensor:
        return self._top_tensor

    def block_hamiltonian(self, site: int) -> Union[Tensor, np.ndarray]:
        """
        Construct the 2-site Hamiltonian from MPO,
        involving both the hopping term and one-site term.

        Args:
            site: The site i.
                The 2-site Hamiltonian is then constructed in site i and i+1.

        Returns:
            M: The 2-site Hamiltonian.
        """
        W1 = self.mpo[site]
        W2 = self.mpo[site + 1]
        if self.n_nodes == 2:
            W1[0] ^ W2[0]
            M = W1 @ W2
        elif site == 0:
            W1[0] ^ W2[0]
            W2[1] ^ self.v_right[0]
            M = W1 @ W2 @ self.v_right
        elif site == self.n_nodes - 2:
            self.v_left[0] ^ W1[0]
            W1[1] ^ W2[0]
            M = self.v_left @ W1 @ W2
        else:
            self.v_left[0] ^ W1[0]
            W1[1] ^ W2[0]
            W2[1] ^ self.v_right[0]
            M = self.v_left @ W1 @ W2 @ self.v_right
        M.reorder_axes([0, 2, 1, 3])
        shape = int(np.sqrt(M.tensor.size))
        return M.tensor.reshape(shape, shape)

    def truncation_gap(self, evals: np.ndarray) -> float:
        """
        Return the gap upon :attr:`~TSDRG.chi` eigenvalues kept.

        Args:
            evals: The eigenvalues (energy spectrum).

        Returns:
            gap: The truncation gap, evals[chi+1] - evals[chi].
        """
        gaps = np.diff(evals)
        return gaps[self.chi - 1] if gaps.size > self.chi else gaps[-1]

    def entanglement_rendering(self, evecs: np.ndarray) -> np.ndarray:
        def von_neumann_entropy(v: np.ndarray) -> float:
            singular_values = np.linalg.svd(v.reshape(2, -1))[1]
            ss = np.square(singular_values)
            return -1 * np.sum(ss @ np.log(ss))
        return np.array([von_neumann_entropy(v) for v in evecs.T])

    def eigen_solver(self, matrix: Union[Tensor, np.ndarray]) -> Tuple[np.ndarray, np.ndarray]:
        evals, evecs = np.linalg.eigh(matrix)
        if matrix.shape[0] > self.chi:
            evals = evals[:self.chi]
            evecs = evecs[:, :self.chi]
        return evals, evecs

    def spectrum_projector(self, site: int, evecs: np.ndarray) -> Tuple[Node, Node]:
        """


        Args:
            site:
            evecs:

        Returns:
            V: The isometric tensor
            W: The coarse-grained MPO
        """
        def network_structure() -> List:
            if self.n_nodes == 2:
                return [(1, 'a1', 'a2'), (1, 'b1', 'b2'), ('a1', 'b1', -3), ('a2', 'b2', -4)]
            elif site == 0:
                return [(1, 'a1', 'a2'), (1, -2, 'b1', 'b2'), ('a1', 'b1', -3), ('a2', 'b2', -4)]
            elif site == self.n_nodes - 2:
                return [(-1, 1, 'a1', 'a2'), (1, 'b1', 'b2'), ('a1', 'b1', -3), ('a2', 'b2', -4)]
            return [(-1, 1, 'a1', 'a2'), (1, -2, 'b1', 'b2'), ('a1', 'b1', -3), ('a2', 'b2', -4)]
        W1 = self.mpo[site].tensor
        W2 = self.mpo[site + 1].tensor
        V = Node(evecs.reshape((W1.shape[-1], W2.shape[-1], evecs.shape[1])))
        W = Node(ncon(
            [W1, W2, V.copy(conjugate=True).tensor, V.tensor],
            network_structure()
        ))
        return V, W

    def neighbouring_bonds(self, bond: int) -> List[int]:
        if bond == 0:
            neighbours = [bond]
        elif bond == self.n_nodes - 1:
            neighbours = [bond - 1]
        else:
            neighbours = [bond - 1, bond]
        return neighbours

    def run(self) -> None:
        """

        Returns:

        """
        evals = self.gap_cache.evals if self.n_nodes == 2 else None
        evecs = self.gap_cache.evecs[0] if self.n_nodes == 2 else None
        for step in count(start=1):
            max_gapped_bond = np.argmax(np.array(self.gap_cache.gap))
            V, W = self.spectrum_projector(max_gapped_bond, self.gap_cache.evecs[max_gapped_bond])
            horizon = self.tree.horizon
            logging.info(f"step {step}, merging TreeNode({horizon[max_gapped_bond]}) "
                         f"and TreeNode({horizon[max_gapped_bond + 1]}) to TreeNode({self.N + step})")
            self._tree.append(
                TreeNode(
                    id=self.N + step - 1,
                    node=V,
                    left=self.tree.node(horizon[max_gapped_bond]),
                    right=self.tree.node(horizon[max_gapped_bond + 1])
                )
            )
            self.mpo.nodes.pop(max_gapped_bond)
            self.mpo.nodes[max_gapped_bond] = W
            self.gap_cache.gap.pop(max_gapped_bond)
            self.gap_cache.evecs.pop(max_gapped_bond)
            if self.n_nodes == 1:
                logging.info('Reach head node of the tree')
                logging.info(f"Obtain ground state energies {evals}")
                self._top_tensor = evecs
                assert step == self.N - 1, "step out of range"
                break
            for bond in self.neighbouring_bonds(max_gapped_bond):
                evals, evecs = self.eigen_solver(self.block_hamiltonian(bond))
                self.gap_cache.gap[bond] = self.truncation_gap(evals)
                self.gap_cache.evecs[bond] = evecs



import numpy as np
from tensornetwork import ncon, Tensor
from tnpy.model import ModelBase
from tnpy.operators import SpinOperators


class RandomHeisenberg(ModelBase):

    def __init__(self, N: int, h: float, penalty: float = 0, s_target: int = 0, trial_id: int = None, seed: int = None):
        """
        Args:
            N: System size.
            h: Disorder strength.
            penalty: Penalty strength (of Lagrangian multiplier).
            s_target: The targeting total Sz charge sector.
            trial_id: ID of the current disorder trial.
            seed: Random seed used to initialize the pseudo-random number generator.
        """
        super(RandomHeisenberg, self).__init__(N)
        self.h = h
        self.penalty = penalty
        self.s_target = s_target
        self.trial_id = trial_id
        self._seed = seed
        rng = np.random.RandomState(self.seed)
        self._random_sequence = rng.uniform(-self.h, self.h, size=self.N)

    def _elem(self, site: int) -> np.ndarray:
        Sp, Sm, Sz, I2, O2 = SpinOperators()

        alpha = self.penalty * (0.25 + self.s_target ** 2 / self.N)
        beta = self._random_sequence[site] - 2.0 * self.penalty * self.s_target

        return np.array(
            [[I2, 0.5 * Sp, 0.5 * Sm, 2.0 * self.penalty * Sz, Sz, alpha * I2 + beta * Sz],
             [O2, O2, O2, O2, O2, Sm],
             [O2, O2, O2, O2, O2, Sp],
             [O2, O2, O2, I2, O2, Sz],
             [O2, O2, O2, O2, O2, Sz],
             [O2, O2, O2, O2, O2, I2]],
            dtype=float
        )

    @property
    def seed(self) -> int:
        return self._seed

    @seed.setter
    def seed(self, seed: int) -> None:
        self._seed = seed
        rng = np.random.RandomState(self.seed)
        self._random_sequence = rng.uniform(-self.h, self.h, size=self.N)


class SpectralFoldedRandomHeisenberg(RandomHeisenberg):

    def __init__(self, N: int, h: float, penalty: float = 0, s_target: int = 0, trial_id: int = None, seed: int = None):
        super(SpectralFoldedRandomHeisenberg, self).__init__(N, h, penalty, s_target, trial_id, seed)

    def _elem(self, site: int) -> Tensor:
        M = super()._elem(site)
        return ncon(
            [M, M],
            [(-1, -3, '-a1', 1), (-2, -4, 1, '-b2')]
        ).reshape((36, 36, 2, 2))



import numpy as np
from collections import namedtuple
from tensornetwork import Node, Tensor, ncon
from typing import Callable, List, Union, Tuple


class SpinOperators:

    def __new__(cls, spin: float = 0.5) -> namedtuple:
        super(SpinOperators, cls).__init__(spin)
        SOp = namedtuple('SpinOperators', ['Sp', 'Sm', 'Sz', 'I2', 'O2'])
        return SOp(Sp=spin * np.array([[0, 2], [0, 0]], dtype=float),
                   Sm=spin * np.array([[0, 0], [2, 0]], dtype=float),
                   Sz=spin * np.array([[1, 0], [0, -1]], dtype=float),
                   I2=np.identity(2, dtype=float),
                   O2=np.zeros((2, 2), dtype=float))


class MPO:

    def __init__(self, N: int, func: Callable):
        self.N = N
        self._nodes = []
        self.nodes = func
        self.__identity = np.identity(self.bond_dimensions)
        self._v_left = Node(self.__identity[0])
        self._v_right = Node(self.__identity[-1])

    def __getitem__(self, site: int):
        return self._nodes[site]

    @property
    def nodes(self) -> List[Node]:
        return self._nodes

    @nodes.setter
    def nodes(self, func: Callable):
        def with_open_boundary(site: int) -> np.ndarray:
            if site == 0:
                return func(site)[0, :, :, :]
            elif site == self.N - 1:
                return func(site)[:, -1, :, :]
            return func(site)
        self._nodes = [Node(with_open_boundary(site)) for site in range(self.N)]

    @property
    def physical_dimensions(self) -> int:
        return self._nodes[0].tensor.shape[-1]

    @property
    def bond_dimensions(self) -> int:
        return self._nodes[0].tensor.shape[0]

    @property
    def v_left(self) -> Node:
        return self._v_left

    @v_left.setter
    def v_left(self, header: int):
        assert header in [0, -1], "header can only accept 0 or -1"
        self._v_left = Node(self.__identity[header])

    @property
    def v_right(self) -> Node:
        return self._v_right

    @v_right.setter
    def v_right(self, header: int):
        assert header in [0, -1], "header can only accept 0 or -1"
        self._v_right = Node(self.__identity[header])


class FullHamiltonian:

    def __init__(self, mpo: MPO):
        self._N = mpo.N
        self._matrix = None
        self.matrix = mpo

    @property
    def N(self) -> int:
        return self._N

    @property
    def matrix(self) -> Union[np.ndarray, Tensor]:
        return self._matrix

    @matrix.setter
    def matrix(self, mpo: MPO):
        def network_structure(site: int) -> Union[Tuple[int, str, str], Tuple[int, int, str, str]]:
            if site == 1:
                return site, f'-a{site}', f'-b{site}'
            elif site == self.N:
                return site-1, f'-a{site}', f'-b{site}'
            return site-1, site, f'-a{site}', f'-b{site}'
        self._matrix = ncon(
            [node.tensor for node in mpo],
            [network_structure(site) for site in range(1, self.N + 1)],
            out_order=[f'-a{site}' for site in range(1, self.N + 1)] +
                      [f'-b{site}' for site in range(1, self.N + 1)]
        )
        self._matrix = self._matrix.reshape(
            (
                mpo.physical_dimensions ** self.N,
                mpo.physical_dimensions ** self.N
            )
        )



import abc
import numpy as np
from tnpy.operators import MPO


class ModelBase(abc.ABC):

    def __init__(self, N: int):
        """

        Args:
            N: System size.
        """
        self.N = N

    @abc.abstractmethod
    def _elem(self, site: int) -> np.ndarray:
        return NotImplemented

    @property
    def mpo(self) -> MPO:
        """
        Return matrix product operator (mpo) as a property of the model.

        Returns:
            mpo:
        """
        return MPO(self.N, self._elem)



import logging
from tensornetwork import Node
from scipy.integrate import solve_ivp
from tnpy.finite_algorithm_base import FiniteAlgorithmBase
from tnpy.linalg import qr
from enum import Enum


class Evolve(Enum):
    FORWARD = 1
    BACKWARD = -1


class FiniteTDVP(FiniteAlgorithmBase):

    def __init__(self, mpo, chi, init_method):
        logging.basicConfig(format='%(asctime)s [%(filename)s] %(levelname)s: %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
        logging.root.setLevel(level=logging.INFO)
        super(FiniteTDVP, self).__init__(mpo, chi, init_method)
        self.center_matrices = {}

    def _unit_solver(self, proceed, t_span, site):

        def forward(t, y):
            M = Node(y.reshape(self.mps_shape(site)))
            W = self.mpo[site]
            if site == 0:
                Renv = self.right_envs[site]
                Rnorm = self.right_norms[site]
                Renv[0] ^ M[2]
                Renv[1] ^ W[0]
                M[1] ^ W[1]
                Renv[2] ^ Rnorm[0]
                result = M @ W @ Renv @ Rnorm
            elif site == self.N-1:
                Lenv = self.left_envs[site]
                Lnorm = self.left_norms[site]
                Lenv[0] ^ M[0]
                Lenv[1] ^ W[0]
                M[1] ^ W[1]
                Lenv[2] ^ Lnorm[0]
                result = Lenv @ Lnorm @ M @ W
            else:
                Lenv = self.left_envs[site]
                Lnorm = self.left_norms[site]
                Renv = self.right_envs[site]
                Rnorm = self.right_norms[site]
                Lenv[0] ^ M[0]
                Lenv[1] ^ W[0]
                M[1] ^ W[2]
                Renv[0] ^ M[2]
                Renv[1] ^ W[1]
                Lenv[2] ^ Lnorm[0]
                Renv[2] ^ Rnorm[0]
                result = Lenv @ Lnorm @ M @ W @ Renv @ Rnorm
            return -1j * result.tensor.reshape(y.shape)

        def backward(t, y):
            C = Node(y.reshape(self.center_matrices[site].tensor.shape))
            Lenv = self.left_envs[site+1]
            Lnorm = self.left_norms[site+1]
            Renv = self.right_envs[site]
            Rnorm = self.right_norms[site]
            Lenv[0] ^ C[0]
            Renv[0] ^ C[1]
            Lenv[1] ^ Renv[1]
            Lenv[2] ^ Lnorm[0]
            Renv[2] ^ Rnorm[0]
            result = Lenv @ Lnorm @ C @ Renv @ Rnorm
            return 1j * result.tensor.reshape(y.shape)

        if proceed == Evolve.FORWARD:
            y0 = self._mps.nodes[site].tensor.reshape(-1).astype(complex)
            result = solve_ivp(forward, t_span, y0)
        elif proceed == Evolve.BACKWARD:
            y0 = self.center_matrices[site].tensor.reshape(-1).astype(complex)
            result = solve_ivp(backward, t_span, y0)
        return result.y[:, -1]

    def sweep(self, iterator, t_span):
        direction = 1 if iterator[0] < iterator[-1] else -1
        for site in iterator:
            theta = self._unit_solver(Evolve.FORWARD, t_span, site)
            if direction == 1:
                theta = theta.reshape(self.d * self.mps_shape(site)[0], -1)
            elif direction == -1:
                theta = theta.reshape(-1, self.d * self.mps_shape(site)[2])
            q, r = qr(theta, chi=self.mps_shape(site)[1+direction])
            if direction == 1:
                self._mps.nodes[site] = Node(q.reshape(self.mps_shape(site)))
                self.center_matrices[site] = Node(r)
                self._update_left_env(site+1)
                self._update_left_norm(site+1)
                if site < self.N-1:
                    C = Node(self._unit_solver(Evolve.BACKWARD, t_span, site).reshape(r.shape))
                    Mp = self._mps.nodes[site+1]
                    C[1] ^ Mp[0]
                    self._mps.nodes[site+1] = C @ Mp
            elif direction == -1:
                self._mps.nodes[site] = Node(r.reshape(self.mps_shape(site)))
                self.center_matrices[site-1] = Node(q)
                self._update_right_env(site-1)
                self._update_right_norm(site-1)
                if site > 0:
                    C = Node(self._unit_solver(Evolve.BACKWARD, t_span, site-1).reshape(q.shape))
                    Mp = self._mps.nodes[site-1]
                    Mp[2] ^ C[0]
            # @TODO: measure something here to check the status of mps
            print(site)
            print(self._mps.check_orthonormality('l', self.N-1))
            print(self._mps.check_orthonormality('r', 0))
            print(self._mps.check_canonical())
            # logging.info("Sweeping to site [{}/{}], norm = ".format(site+1, self.N))
        return

    def evolve(self, t_span):
        pass
        #



import numpy as np
from tnpy.model import ModelBase
from tnpy.operators import SpinOperators


class DimerXXZ(ModelBase):

    def __init__(self, N: int, J: float, delta: float, h: float,
                 penalty: float = 0, s_target: int = 0, trial_id: int = None):
        """
        Args:
            N: System size.
            J:
            delta:
            h: Disorder strength.
            penalty: Penalty strength (or Lagrangian multiplier).
            s_target: The targeting total Sz charge sector.
            trial_id: ID of the current disorder trial.
        """
        super(DimerXXZ, self).__init__(N)
        self.J = J
        self.delta = delta
        self.h = h
        self.penalty = penalty
        self.s_target = s_target
        self.trial_id = trial_id

    def _elem(self, site: int) -> np.ndarray:
        Sp, Sm, Sz, I2, O2 = SpinOperators(spin=1)

        rand_J = (1 + self.delta * (-1) ** site) * np.random.uniform() ** self.J
        alpha = self.penalty * (0.25 + self.s_target ** 2 / self.N)
        beta = np.random.uniform(-self.h, self.h) - 2.0 * self.penalty * self.s_target

        return np.array(
            [[I2, 0.5 * rand_J * Sp, 0.5 * rand_J * Sm, 2.0 * self.penalty * Sz, Sz, alpha * I2 + beta * Sz],
             [O2, O2, O2, O2, O2, Sm],
             [O2, O2, O2, O2, O2, Sp],
             [O2, O2, O2, I2, O2, Sz],
             [O2, O2, O2, O2, O2, Sz],
             [O2, O2, O2, O2, O2, I2]],
            dtype=float
        )



from setuptools import setup
# from distutils.core import setup
from tnpy import __version__


with open("README.md", "r") as f:
    long_description = f.read()

with open('requirements.txt') as f:
    required = f.read().splitlines()


setup(
    name='tnpy',
    packages=['tnpy', 'tnpy.model'],
    package_dir={'tnpy': 'tnpy', 'tnpy.model': 'tnpy/model'},
    version=__version__,
    license='Apache 2.0',
    description='Tensor Network Algorithms',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='tao-lin',
    author_email='tanlin2013@gmail.com',
    url='https://github.com/tanlin2013/tnpy',
    download_url=f'https://github.com/tanlin2013/tnpy/archive/v{__version__}.tar.gz',
    keywords=['mps', 'tensor network'],
    install_requires=required,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Mathematics',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'
    ],
)
